/*
** mod_anonip v2.1.1
**
** An Apache2 module to anonymize the IP address of any request by a
** configurable number of bits
**
** Copyright (C) 17.6.2016 by Andreas Ley <andreas.ley@kit.edu>
**
** A complete rewrite, but based on an earlier version by Torben Haase
** <Torben.Letorbi@gmail.com> and Malte Starostik <M-Starostik@versanet.de>
** which was based on mod_removeip by Andrew McNaughton <Andrew@scoop.co.nz>
** and inspired by mod_log_ipmask by webfactory GmbH (webfactory.de) which
** in turn is a fork of the version by Saxonia's data protection officer
** (saechsdsb.de/ipmask).
**
** This program is free software: you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation, either version 3 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

static char copyright[] = "@(#)Copyright (C) 2016-2017 by Andreas Ley (andreas.ley@kit.edu)";
static char sccsid[] = "@(#)mod_anonip 2.1.1 - An Apache2 module to anonymize the IP address of any request by a configurable number of bits";

/*
#define ANONIP_LEGACY
#define ANONIP_DIR_CONFIG
*/
#define _ANONIP_DEBUG																	/*@DEBUG@*/
#define _ANONIP_DEBUG_ALL																/*@DEBUG@*/

#include "apr_strings.h"
/* What did we need this for?!
#include "apr_network_io.h"
*/

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_protocol.h"
/* What did we need this for?!
#include "http_request.h"
*/
#include "mod_log_config.h"

module AP_MODULE_DECLARE_DATA anonip_module;

typedef struct {
#if APR_HAVE_IPV6
	apr_uint32_t	maskv6[4];
#endif
	apr_uint32_t	maskv4[1];
	unsigned int	mode:2;
#if APR_HAVE_IPV6
	unsigned int	maskv6_set:1;
#endif
	unsigned int	maskv4_set:1;
	unsigned int	mode_set:1;
} anonip_config;

/* Stolen from apr/network_io/unix/sockaddr.c
 * Unfortunatly the struct isn't exported, but we want to take advantage of apr_ipsubnet_create to convert a masklen to a bitmask
 * Still seeking for an 'official' function to do this conversion without replicating code :-(
 */
struct apr_ipsubnet_t {
	int		family;
#if APR_HAVE_IPV6
	apr_uint32_t	sub[4]; /* big enough for IPv4 and IPv6 addresses */
	apr_uint32_t	mask[4];
#else
	apr_uint32_t	sub[1];
	apr_uint32_t	mask[1];
#endif
};

static void *anonip_create_config(apr_pool_t *p) {
	anonip_config	*config = (anonip_config *)apr_palloc(p, sizeof(anonip_config));

#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_create_config config: %08lx\n",(long unsigned int)config);										/*@DEBUG@*/
#endif																			/*@DEBUG@*/
#if APR_HAVE_IPV6
	config->maskv6[0] = 0xFFFFFFFF;
	config->maskv6[1] = 0xFFFFFFFF;
	config->maskv6[2] = 0xFFFFFFFF;
	config->maskv6[3] = 0xFFFFFFFF;
	config->maskv6_set = 0;
#ifdef ANONIP_DEBUG_ALL																	/*@DEBUG@*/
printf("init maskv6: %08x %08x %08x %08x\n",config->maskv6[0],config->maskv6[1],config->maskv6[2],config->maskv6[3]);					/*@DEBUG@*/
#endif																			/*@DEBUG@*/
#endif

	config->maskv4[0] = 0xFFFFFFFF;
	config->maskv4_set = 0;
#ifdef ANONIP_DEBUG_ALL																	/*@DEBUG@*/
printf("init maskv4: %08x\n",config->maskv4[0]);													/*@DEBUG@*/
#endif																			/*@DEBUG@*/

	config->mode = 1;
	config->mode_set = 0;
#ifdef ANONIP_DEBUG_ALL																	/*@DEBUG@*/
printf("init mode: %d\n",config->mode);															/*@DEBUG@*/
#endif																			/*@DEBUG@*/

	return (void *)config;
}

/* For the sake of avoiding incompatible pointer type warnings */
static void *anonip_create_dir_config(apr_pool_t *p, char *dir) {
	return anonip_create_config(p);
}

static void *anonip_create_server_config(apr_pool_t *p, server_rec *s) {
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_create_server_config server_hostname: %s\n", s->server_hostname);									/*@DEBUG@*/
server_addr_rec *a = s->addrs; while (a) { char *s; apr_sockaddr_ip_get(&s,a->host_addr); printf("                                       addr: %s:%d\n",s,a->host_port); a = a->next; }																		/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	return anonip_create_config(p);
}

static void *anonip_merge_config(apr_pool_t *p, void *basev, void *addv) {
	anonip_config *base = (anonip_config *)basev;
	anonip_config *add = (anonip_config *)addv;
	anonip_config *new = (anonip_config *)apr_palloc(p, sizeof(anonip_config));

#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_merge_config config: %08lx + %08lx -> %08lx\n", (long unsigned int)base, (long unsigned int)add, (long unsigned int)new);		/*@DEBUG@*/
#endif																			/*@DEBUG@*/
#if APR_HAVE_IPV6
#ifdef ANONIP_DEBUG_ALL																	/*@DEBUG@*/
printf("base maskv6: %08x %08x %08x %08x\n",base->maskv6[0],base->maskv6[1],base->maskv6[2],base->maskv6[3]);						/*@DEBUG@*/
printf("add maskv6: %08x %08x %08x %08x\n",add->maskv6[0],add->maskv6[1],add->maskv6[2],add->maskv6[3]);						/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	new->maskv6[0] = add->maskv6_set ? add->maskv6[0] : base->maskv6[0];
	new->maskv6[1] = add->maskv6_set ? add->maskv6[1] : base->maskv6[1];
	new->maskv6[2] = add->maskv6_set ? add->maskv6[2] : base->maskv6[2];
	new->maskv6[3] = add->maskv6_set ? add->maskv6[3] : base->maskv6[3];
	new->maskv6_set = add->maskv6_set || base->maskv6_set;
#ifdef ANONIP_DEBUG_ALL																	/*@DEBUG@*/
printf("merge maskv6: %08x %08x %08x %08x\n",new->maskv6[0],new->maskv6[1],new->maskv6[2],new->maskv6[3]);						/*@DEBUG@*/
#endif																			/*@DEBUG@*/
#endif

	new->maskv4[0] = add->maskv4_set ? add->maskv4[0] : base->maskv4[0];
	new->maskv4_set = add->maskv4_set || base->maskv4_set;
#ifdef ANONIP_DEBUG_ALL																	/*@DEBUG@*/
printf("merge maskv4: %08x\n",new->maskv4[0]);														/*@DEBUG@*/
#endif																			/*@DEBUG@*/

	new->mode = add->mode_set ? add->mode : base->mode;
	new->mode_set = add->mode_set || base->mode_set;
#ifdef ANONIP_DEBUG_ALL																	/*@DEBUG@*/
printf("merge mode: %d\n",new->mode);														/*@DEBUG@*/
#endif																			/*@DEBUG@*/

	return (void *)new;
}

#if APR_HAVE_IPV6
static const char *anonip_maskv6(cmd_parms *cmd, void *conf, const char* mask) {
	apr_status_t rv;
	apr_ipsubnet_t *ip;

	rv = apr_ipsubnet_create(&ip, "::", mask, cmd->pool);
	if (rv != APR_SUCCESS)
		return apr_psprintf(cmd->pool, "%s: Invalid mask %s: %pm", cmd->cmd->name, mask, &rv);

#ifdef ANONIP_DIR_CONFIG
	anonip_config *config = (anonip_config *)conf;
#else
	anonip_config *config = ap_get_module_config(cmd->server->module_config, &anonip_module);
#endif
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_maskv6 config: %08lx \"%s\"\n",(long unsigned int)config,mask);										/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	config->maskv6[0] = ip->mask[0];
	config->maskv6[1] = ip->mask[1];
	config->maskv6[2] = ip->mask[2];
	config->maskv6[3] = ip->mask[3];
	config->maskv6_set = 1;
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("       config maskv6: %08x %08x %08x %08x\n",config->maskv6[0],config->maskv6[1],config->maskv6[2],config->maskv6[3]);			/*@DEBUG@*/
#endif			/*@DEBUG@*/
	return NULL;
}
#endif

static const char *anonip_maskv4(cmd_parms *cmd, void *conf, const char* mask) {
	apr_status_t rv;
	apr_ipsubnet_t *ip;

	rv = apr_ipsubnet_create(&ip, "0.0.0.0", mask, cmd->pool);
	if (rv != APR_SUCCESS)
		return apr_psprintf(cmd->pool, "%s: Invalid mask %s: %pm", cmd->cmd->name, mask, &rv);

#ifdef ANONIP_DIR_CONFIG
	anonip_config *config = (anonip_config *)conf;
#else
	anonip_config *config = ap_get_module_config(cmd->server->module_config, &anonip_module);
#endif
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_maskv4 config: %08lx \"%s\"\n",(long unsigned int)config,mask);										/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	config->maskv4[0] = ip->mask[0];
	config->maskv4_set = 1;
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("       config maskv4: %08x\n",config->maskv4[0]);											/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	return NULL;
}

#ifdef ANONIP_LEGACY
static const char *anonip_mask(cmd_parms *cmd, void *conf, const char* mask) {
	const char *err;

#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_mask conf: %08lx \"%s\"\n",(long unsigned int)conf,mask);										/*@DEBUG@*/
#endif																			/*@DEBUG@*/
#if APR_HAVE_IPV6
	err = anonip_maskv6(cmd, conf, apr_psprintf(cmd->pool, "%d", 128-16*atoi(mask)));
	if (err == NULL)
#endif
		err = anonip_maskv4(cmd, conf, apr_psprintf(cmd->pool, "%d", 32-8*atoi(mask)));
	if (err != NULL)
		return apr_psprintf(cmd->pool, "%s: Invalid byte mask %s", cmd->cmd->name, mask);
	return NULL;
}
#endif

static const char *anonip_mode(cmd_parms *cmd, void *conf, const char* mode) {

#ifdef ANONIP_DIR_CONFIG
	anonip_config *config = (anonip_config *)conf;
#else
	anonip_config *config = ap_get_module_config(cmd->server->module_config, &anonip_module);
#endif
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_mode config: %08lx \"%s\"\n",(long unsigned int)config,mode);										/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	if (!strcasecmp(mode,"none")) {
		config->mode = 0;
		config->mode_set = 1;
	}
	else if (!strcasecmp(mode,"all")) {
		config->mode = 1;
		config->mode_set = 1;
	}
	else if (!strcasecmp(mode,"text")) {
		config->mode = 2;
		config->mode_set = 1;
	}
	else if (!strcasecmp(mode,"log")) {
		config->mode = 3;
		config->mode_set = 1;
	}
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("       config mode: %d\n",config->mode);													/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	return NULL;
}

void anonip_mask_ip(anonip_config *config, apr_sockaddr_t *sa, char **ip, int text) {
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_mask_ip	   sa: %08lx\n",(long unsigned int)sa);												/*@DEBUG@*/
printf("                 text: %d\n",text);													/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	apr_sockaddr_t sa_text;
	apr_uint32_t ipaddr_text[4];
	if (text) {
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("          ipaddr_text: %08lx\n",(long unsigned int)&ipaddr_text);										/*@DEBUG@*/
#endif																			/*@DEBUG@*/
		memcpy(&sa_text,sa,sizeof(sa_text));
		/* Assert sa->ipaddr_len <= sizeof(ipaddr_text) */
		memcpy(&ipaddr_text,sa->ipaddr_ptr,sa->ipaddr_len);
		sa_text.ipaddr_ptr = &ipaddr_text;
		sa = &sa_text;
	}
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("                   sa: %08lx\n",(long unsigned int)sa);												/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	apr_uint32_t *addr = (apr_uint32_t *)sa->ipaddr_ptr;
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("                 addr: %08lx\n",(long unsigned int)addr);												/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	int mask;

#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_mask_ip config: %08lx\n",(long unsigned int)config);											/*@DEBUG@*/
printf("                 addr: %08x",addr[0]);														/*@DEBUG@*/
#if APR_HAVE_IPV6																	/*@DEBUG@*/
if (sa->family == AF_INET6)																/*@DEBUG@*/
printf(" %08x %08x %08x",addr[1],addr[2],addr[3]);													/*@DEBUG@*/
#endif																			/*@DEBUG@*/
printf("\n");																		/*@DEBUG@*/
#endif																			/*@DEBUG@*/
#if APR_HAVE_IPV6
	if (sa->family == AF_INET6) {
		if (IN6_IS_ADDR_V4MAPPED((struct in6_addr *)sa->ipaddr_ptr))
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
{																			/*@DEBUG@*/
#endif																			/*@DEBUG@*/
			addr[3] &= config->maskv4[0];
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_mask_ip mapped: %08x %08x %08x %08x\n",0xFFFFFFFF,0xFFFFFFFF,0xFFFFFFFF,config->maskv4[0]);						/*@DEBUG@*/
}
#endif			/*@DEBUG@*/
		else {
			addr[0] &= config->maskv6[0];
			addr[1] &= config->maskv6[1];
			addr[2] &= config->maskv6[2];
			addr[3] &= config->maskv6[3];
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_mask_ip maskv6: %08x %08x %08x %08x\n",config->maskv6[0],config->maskv6[1],config->maskv6[2],config->maskv6[3]);				/*@DEBUG@*/
#endif			/*@DEBUG@*/
		}
	}
	else
#endif
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
{																			/*@DEBUG@*/
#endif																			/*@DEBUG@*/
		addr[0] &= config->maskv4[0];
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_mask_ip maskv4: %08x\n",config->maskv4[0]);												/*@DEBUG@*/
}																			/*@DEBUG@*/
#endif																			/*@DEBUG@*/

#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_mask_ip masked: %08x",addr[0]);														/*@DEBUG@*/
#if APR_HAVE_IPV6																	/*@DEBUG@*/
if (sa->family == AF_INET6)																/*@DEBUG@*/
printf(" %08x %08x %08x",addr[1],addr[2],addr[3]);													/*@DEBUG@*/
#endif																			/*@DEBUG@*/
printf("\n");																		/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	apr_sockaddr_ip_get(ip,sa);
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("IP: %s\n",*ip);																	/*@DEBUG@*/
#endif																			/*@DEBUG@*/
}

static int anonip_hook(request_rec *r) {
#ifdef ANONIP_DIR_CONFIG
	anonip_config *config = (anonip_config *)ap_get_module_config(r->per_dir_config, &anonip_module);
#else
	anonip_config *config = (anonip_config *)ap_get_module_config(r->server->module_config, &anonip_module);
#endif
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_hook config: %08lx\n",(long unsigned int)config);											/*@DEBUG@*/
printf("       config mode: %d\n",config->mode);													/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	if (config->mode == 1 || config->mode == 2) {
		conn_rec *c = r->connection;

#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
apr_uint32_t *addr = (apr_uint32_t *)c->client_addr->ipaddr_ptr;											/*@DEBUG@*/
printf("			  addr: %08x",addr[0]);													/*@DEBUG@*/
#if APR_HAVE_IPV6																	/*@DEBUG@*/
if (c->client_addr->family == AF_INET6)															/*@DEBUG@*/
printf(" %08x %08x %08x",addr[1],addr[2],addr[3]);													/*@DEBUG@*/
#endif																			/*@DEBUG@*/
printf("\n");																		/*@DEBUG@*/
#endif																			/*@DEBUG@*/

#if AP_SERVER_MAJORVERSION_NUMBER > 2 || (AP_SERVER_MAJORVERSION_NUMBER == 2 && AP_SERVER_MINORVERSION_NUMBER >= 4)
		anonip_mask_ip(config, c->client_addr, &c->client_ip, config->mode == 2);
		anonip_mask_ip(config, r->useragent_addr, &r->useragent_ip, config->mode == 2);
#else
		anonip_mask_ip(config, c->remote_addr, &c->remote_ip, config->mode == 2);
#endif
	}

	return DECLINED;
}

static const char *log_remote_address_masked(request_rec* r, char* mask)
{
#ifdef ANONIP_DIR_CONFIG
	anonip_config *config = (anonip_config *)ap_get_module_config(r->per_dir_config, &anonip_module);
#else
	anonip_config *config = (anonip_config *)ap_get_module_config(r->server->module_config, &anonip_module);
#endif
	conn_rec *c = r->connection;

#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("log_remote_address_masked config: %08lx\n",(long unsigned int)config);										/*@DEBUG@*/
printf("                     config mode: %d\n",config->mode);												/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	if (config->mode == 3) {
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
apr_uint32_t *addr = (apr_uint32_t *)c->client_addr->ipaddr_ptr;											/*@DEBUG@*/
printf("                            addr: %08x",addr[0]);												/*@DEBUG@*/
#if APR_HAVE_IPV6																	/*@DEBUG@*/
if (c->client_addr->family == AF_INET6)															/*@DEBUG@*/
printf(" %08x %08x %08x",addr[1],addr[2],addr[3]);													/*@DEBUG@*/
#endif																			/*@DEBUG@*/
printf("\n");																		/*@DEBUG@*/
#endif																			/*@DEBUG@*/
		char* ip;
#if AP_SERVER_MAJORVERSION_NUMBER > 2 || (AP_SERVER_MAJORVERSION_NUMBER == 2 && AP_SERVER_MINORVERSION_NUMBER >= 4)
		anonip_mask_ip(config, mask && strcmp(mask, "c") ? r->useragent_addr : c->client_addr, &ip, 1);
#else
		anonip_mask_ip(config, c->remote_addr, &ip, 1);
#endif
		return(ip);
	}

#if AP_SERVER_MAJORVERSION_NUMBER > 2 || (AP_SERVER_MAJORVERSION_NUMBER == 2 && AP_SERVER_MINORVERSION_NUMBER >= 4)
	return(mask && strcmp(mask, "c")?r->useragent_ip:c->client_ip);
#else
	return(c->remote_ip);
#endif
}

static const char *log_remote_host_masked(request_rec* r, char* mask)
{
#ifdef ANONIP_DIR_CONFIG
	anonip_config *config = (anonip_config *)ap_get_module_config(r->per_dir_config, &anonip_module);
#else
	anonip_config *config = (anonip_config *)ap_get_module_config(r->server->module_config, &anonip_module);
#endif
	conn_rec *c = r->connection;

#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("log_remote_host_masked config: %08lx\n",(long unsigned int)config);										/*@DEBUG@*/
printf("                  config mode: %d\n",config->mode);												/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	if (config->mode == 3) {
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
apr_uint32_t *addr = (apr_uint32_t *)c->client_addr->ipaddr_ptr;											/*@DEBUG@*/
printf("                         addr: %08x",addr[0]);												/*@DEBUG@*/
#if APR_HAVE_IPV6																	/*@DEBUG@*/
if (c->client_addr->family == AF_INET6)															/*@DEBUG@*/
printf(" %08x %08x %08x",addr[1],addr[2],addr[3]);													/*@DEBUG@*/
#endif																			/*@DEBUG@*/
printf("\n");																		/*@DEBUG@*/
#endif																			/*@DEBUG@*/
		char* ip;
#if AP_SERVER_MAJORVERSION_NUMBER > 2 || (AP_SERVER_MAJORVERSION_NUMBER == 2 && AP_SERVER_MINORVERSION_NUMBER >= 4)
		anonip_mask_ip(config, c->client_addr, &ip, 1);
#else
		anonip_mask_ip(config, c->remote_addr, &ip, 1);
#endif
		return(ip);
	}

	return (ap_escape_logitem(r->pool, ap_get_remote_host(c, r->per_dir_config, REMOTE_NAME, NULL)));
}

#if AP_SERVER_MAJORVERSION_NUMBER > 2 || (AP_SERVER_MAJORVERSION_NUMBER == 2 && AP_SERVER_MINORVERSION_NUMBER >= 4)
static int core_log_remote_address_masked(const ap_errorlog_info *info, const char *mask, char *buf, int buflen)
{
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("core_log_remote_address_masked\n");														/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	const conn_rec *c = info->c;
	if (!c)
		return(0);

	const request_rec *r = info->r;
	apr_sockaddr_t *sa;
	char *ip;

	if (r && mask && strcmp(mask, "c")) {
		sa = r->useragent_addr;
		ip = r->useragent_ip;
	}
	else {
		sa = c->client_addr;
		ip = c->client_ip;
	}

	if (r) {
#ifdef ANONIP_DIR_CONFIG
		anonip_config *config = (anonip_config *)ap_get_module_config(r->per_dir_config, &anonip_module);
#else
		anonip_config *config = (anonip_config *)ap_get_module_config(r->server->module_config, &anonip_module);
#endif

#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("core_log_remote_address_masked config: %08lx\n",(long unsigned int)config);									/*@DEBUG@*/
printf("                          config mode: %d\n",config->mode);											/*@DEBUG@*/
#endif																			/*@DEBUG@*/
		if (config->mode == 3) {
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
apr_uint32_t *addr = (apr_uint32_t *)sa->ipaddr_ptr;													/*@DEBUG@*/
printf("                                 addr: %08x",addr[0]);												/*@DEBUG@*/
#if APR_HAVE_IPV6																	/*@DEBUG@*/
if (sa->family == AF_INET6)																/*@DEBUG@*/
printf(" %08x %08x %08x",addr[1],addr[2],addr[3]);													/*@DEBUG@*/
#endif																			/*@DEBUG@*/
printf("\n");																		/*@DEBUG@*/
#endif																			/*@DEBUG@*/
			anonip_mask_ip(config, sa, &ip, 1);
		}
	}
	return(apr_snprintf(buf, buflen, "%s:%d", ip, sa ? sa->port : 0));
}
#endif

static const command_rec anonip_cmds[] = {
#if APR_HAVE_IPV6
	AP_INIT_TAKE1(
		"AnonipMaskV6",
		anonip_maskv6,
		NULL,
#ifdef ANONIP_DIR_CONFIG
		ACCESS_CONF |
#endif
		RSRC_CONF,
		"IPv6 bitmask (Show this many bits of the IPv6 address, defaults to 128)"
		),
#endif
	AP_INIT_TAKE1(
		"AnonipMaskV4",
		anonip_maskv4,
		NULL,
#ifdef ANONIP_DIR_CONFIG
		ACCESS_CONF |
#endif
		RSRC_CONF,
		"IPv4 bitmask (Show this many bits of the IPv4 address, defaults to 32)"
		),
#ifdef ANONIP_LEGACY
	AP_INIT_TAKE1(
		"AnonipMask",
		anonip_mask,
		NULL,
#ifdef ANONIP_DIR_CONFIG
		ACCESS_CONF |
#endif
		RSRC_CONF,
		"(DEPRECATED) Set the bytemask for mod_anonip (e.g. 0=192.168.1.1, 3=192.0.0.0)"
		),
#endif
	AP_INIT_TAKE1(
		"AnonipMode",
		anonip_mode,
		NULL,
#ifdef ANONIP_DIR_CONFIG
		ACCESS_CONF |
#endif
		RSRC_CONF,
		"Specify anonymization mode (none, all, text or log, defaults to all)"
		),
	{ NULL }
};

static int anonip_pre_config(apr_pool_t* p, apr_pool_t* plog, apr_pool_t* ptemp)
{
	static APR_OPTIONAL_FN_TYPE(ap_register_log_handler) *anonip_pfn_register;

	anonip_pfn_register = APR_RETRIEVE_OPTIONAL_FN(ap_register_log_handler);
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_pre_config anonip_pfn_register: %08lx\n",(long unsigned int)anonip_pfn_register);							/*@DEBUG@*/
#endif																			/*@DEBUG@*/
	if (anonip_pfn_register) {
		anonip_pfn_register(p, "a", log_remote_address_masked, 0);
		anonip_pfn_register(p, "h", log_remote_host_masked, 0);
	}

	return OK;
}

static void anonip_register_hooks(apr_pool_t *p) {
	static const char* const log_config[] = {"mod_log_config.c", NULL};

#if AP_SERVER_MAJORVERSION_NUMBER > 2 || (AP_SERVER_MAJORVERSION_NUMBER == 2 && AP_SERVER_MINORVERSION_NUMBER >= 4)
	ap_register_errorlog_handler(p, "a", core_log_remote_address_masked, 0);
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("core_log_remote_address_masked registered\n");													/*@DEBUG@*/
#endif																			/*@DEBUG@*/
#endif
	ap_hook_pre_config(anonip_pre_config, log_config, NULL, APR_HOOK_FIRST);
	/*
	ap_hook_pre_config(anonip_pre_config, NULL, NULL, APR_HOOK_REALLY_FIRST);
	*/
#ifdef ANONIP_DIR_CONFIG
	/* Final per directory configuration isn't available until this stage */
	ap_hook_header_parser(anonip_hook, NULL, NULL, APR_HOOK_MIDDLE);
#else
	/* Per server configuration can be applied this early */
	ap_hook_post_read_request(anonip_hook, NULL, NULL, APR_HOOK_MIDDLE);
#endif
#ifdef ANONIP_DEBUG																	/*@DEBUG@*/
printf("anonip_register_hooks\n");															/*@DEBUG@*/
#endif																			/*@DEBUG@*/
}

module AP_MODULE_DECLARE_DATA anonip_module = {
	STANDARD20_MODULE_STUFF,
#ifdef ANONIP_DIR_CONFIG
	anonip_create_dir_config,
	anonip_merge_config,
	NULL,
	NULL,
#else
	NULL,
	NULL,
	anonip_create_server_config,
	anonip_merge_config,
#endif
	anonip_cmds,
	anonip_register_hooks,
};
